/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "export_utils.h"
#include "scripture.h"
#include "book.h"
#include "bible.h"
#include "usfm.h"
#include "usfmtools.h"
#include "project.h"
#include "notes_utils.h"
#include "date_time_utils.h"
#include "export_translation_notes.h"
#include "bookmetrics.h"
#include "constants.h"
#include "gwrappers.h"
#include "categorize_chapter.h"
#include "dialogsword.h"
#include "directories.h"
#include "unixwrappers.h"
#include "gtkwrappers.h"
#include "shell.h"


void export_to_bibleworks (Configuration * configuration, GtkWidget * parent, Progress * progress)
/*
Exports a whole project to a file that is fit for being imported by the 
BibleWorks Version Database Compiler.
This done in this manner:
- If an \id line is found, use its ID to get the name of the book in BibleWorks.
- If an chapter tag if found, use its number.
- As soon as an ID is found, put the current chapter at one.
- If a verse is found, add its number.
- Text of tags known not to be printable is thrown out.
- Any other tag's text gets added to the text of the verse.
Yes, this is a bit rough, I know...
*/
{
  // Get the filename where the data to save to.
  GtkWidget *dialog;
  dialog = gtk_file_chooser_dialog_new ("Save to a file", GTK_WINDOW (parent),
    GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
    GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (dialog), configuration->export_to_bibleworks_filename.c_str ());
  int result = gtk_dialog_run (GTK_DIALOG (dialog));
  ustring filename;
  if (result == GTK_RESPONSE_ACCEPT) 
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
  gtk_widget_destroy (dialog);  
  if (result == GTK_RESPONSE_ACCEPT) {
    // Store (new) filename.
    configuration->export_to_bibleworks_filename = filename;
    // Notebook shows progress bar.
    progress->start();
    // Start process.
    try
    {
      Project project (configuration->project);
      Usfm usfm (project.get_stylesheet());
      // Write to outputfile.
      WriteText wt (filename);
      // Open the Scripture and go through it.
      Scripture scripture (configuration->project);
      progress->set_iterate (0, 1, scripture.books.size());
      progress->set_text ("Exporting");
      for (unsigned int bk = 0; bk < scripture.books.size(); bk++) {
        // Progress information.
        progress->iterate();
        if (progress->cancel) {
          progress->finish();
          return;
        }
        // Get the abbreviation to be used.
        ustring abbreviation;
        unsigned index = english_id_to_index (scripture.books[bk]);
        abbreviation = index_to_bibleworks_id (index);
        // Open the book and its metrics.
        Book book (scripture.paths[bk], false, "");
        BookMetrics bookmetrics (scripture.paths[bk]);
        // Handle each chapter in the book.
        vector<unsigned int> chapters = bookmetrics.get_chapters();
        for (unsigned int ch = 0; ch < chapters.size(); ch++) {
          // Do not export chapter 0.
          if (chapters[ch] == 0)
            continue;
          // Get the chapter and handle it.
          vector<ustring> lines;
          book.get_chapter (chapters[ch], lines);
          CategorizeChapter categorize_chapter (lines);
          for (unsigned int vs = 0; vs < categorize_chapter.verses.size(); vs++) {
            // Do not export verse 0.
            if (categorize_chapter.verses[vs] == "0")
              continue;
            // Output verse.
            wt.text(abbreviation);
            wt.text(" ");
            wt.text(convert_to_string (chapters[ch]));
            wt.text(":");
            wt.text(categorize_chapter.verse_texts[vs]);
            wt.text("\n");
          }
        }
      }
    }
    catch (exception & ex)
    {
      cerr << "Converting to BibleWorks: " << ex.what () << endl;
    }
  }
  // Notebook hides progress bar.
  progress->finish();
}


void export_translation_notes (const ustring& filename, ExportNotesFormat format, const vector<gint> & ids_to_display, bool export_all, GtkWidget * parent, Progress * progress)
{
  ExportTranslationNotes etn (filename, format, ids_to_display, export_all, parent, progress);
}


void export_to_sword (Configuration * configuration, GtkWidget * parent, Progress * progress)
/*
Exports a whole project to a SWORD module.
At the time of writing this, the information on how to create a module for 
sword was found at http://www.crosswire.org/sword/develop/swordmodule/
Here's how we do the conversion
- A .conf file is created and put in ~/.sword/mods.d/
- We use the "imp" format as this is easiest.
- Create the module with program imp2vs.
- The text will be stored in ~/.sword/modules/texts/bibledit/<name>/
*/
{
  // Dialog for information entry.
  ustring module_name;
  {
    SwordDialog sworddialog (configuration);
    if (sworddialog.run() != GTK_RESPONSE_OK)
      return;
    module_name = sworddialog.name;
  }
  // Check for converter.
  if (!unix_which ("imp2vs")) {
    gtkw_dialog_error (NULL, "The SWORD compiler was not found.");
    return;
  }
  // Notebook shows progress bar.
  progress->start();
  progress->set_text ("Preparing");
  // The temporal directories for the data.
  ustring base_directory = gw_build_filename (directories_get_temp(), "sword");
  remove_directory (base_directory);
  create_directory (base_directory);
  ustring absolute_conf_directory = gw_build_filename (base_directory, "mods.d");
  create_directory (absolute_conf_directory);
  ustring relative_text_directory = gw_build_filename ("modules", "texts", "bibledit", lowerCase (module_name));
  ustring absolute_text_directory = gw_build_filename (base_directory, relative_text_directory);
  create_directory (absolute_text_directory);
  // Open the project that contains the data.
  Project project (configuration);
  // Create the configuration file.
  vector<ustring> lines;
  ustring line;
  lines.push_back ("[" + module_name + "]");
  line = "DataPath=.";
  line.append (G_DIR_SEPARATOR_S + relative_text_directory + G_DIR_SEPARATOR_S);
  lines.push_back (line);
  lines.push_back ("ModDrv=RawText");
  lines.push_back ("BlockType=CHAPTER");
  lines.push_back ("SourceType=GBF");
  lines.push_back ("Lang=" + project.sword_language);
  lines.push_back ("Version=" + project.sword_version);
  lines.push_back ("Encoding=UTF-8");
  lines.push_back ("Description=" + project.sword_description);
  lines.push_back ("About=" + project.sword_about);
  lines.push_back ("LCSH=" + project.sword_lcsh);
  lines.push_back ("DistributionLicense=" + project.sword_license);
  write_lines (gw_build_filename (absolute_conf_directory, lowerCase (module_name) + ".conf"), lines);
  lines.clear();
  // Start process of producing the text file.
  ustring inputfile;
  try
  {
    Usfm usfm (project.get_stylesheet());
    // Write to inputfile.
    inputfile = gw_build_filename (directories_get_temp(), lowerCase (module_name) + ".input");
    WriteText wt (inputfile);
    // Open the Scripture and go through it.
    Scripture scripture (configuration->project);
    progress->set_iterate (0, 1, scripture.books.size());
    for (unsigned int bk = 0; bk < scripture.books.size(); bk++) {
      // Progress information.
      progress->iterate();
      if (progress->cancel) {
        progress->finish();
        return;
      }
      // Get the abbreviation to be used.
      ustring abbreviation;
      unsigned index = english_id_to_index (scripture.books[bk]);
      abbreviation = index_to_osis_id (index);
      // Open the book and its metrics.
      Book book (scripture.paths[bk], false, "");
      BookMetrics bookmetrics (scripture.paths[bk]);
      // Deal with each chapter in the book.
      vector<unsigned int> chapters = bookmetrics.get_chapters();
      for (unsigned int ch = 0; ch < chapters.size(); ch++) {
        // Do not export chapter 0.
        if (chapters[ch] == 0)
          continue;
        // Get the chapter and handle it.
        vector<ustring> lines;
        book.get_chapter (chapters[ch], lines);
        CategorizeChapter categorize_chapter (lines);
        for (unsigned int vs = 0; vs < categorize_chapter.verses.size(); vs++) {
          // Do not export verse 0.
          if (categorize_chapter.verses[vs] == "0")
            continue;
          // Write the tag.
          wt.text ("$$$");
          wt.text (abbreviation + " " + convert_to_string (chapters[ch]) + ":" + categorize_chapter.verses[vs]);
          wt.text ("\n");
          // Remove the verse number from the text.
          ustring text (categorize_chapter.verse_texts[vs]);
          size_t pos;
          pos = text.find (" ");
          if (pos != string::npos) {
            text.erase (0, ++pos);
          }
          // Write the text.
          wt.text (text);
          wt.text ("\n");
        }
      }
    }
  }
  catch (exception & ex)
  {
    cerr << "Export: " << ex.what () << endl;
  }
  // Convert the inputfile using the sword api utility.
  ustring command;
  command = "imp2vs " + inputfile + " " + absolute_text_directory;
  run_shell_progress (command, "imp2vs", progress, "Compiling", true, 3000000);
  // Install it.
  command = "cp -r " + gw_build_filename (base_directory, "*") + " " + configuration->export_to_sword_install_path;
  system (command.c_str());
  // Compress and save the module.
  ustring zipfile = temporary_file (module_name) + ".zip";
  unlink (zipfile.c_str());
  command = "cd '" + base_directory + "'; ";
  command.append ("zip -r '" + zipfile + "' *");
  system (command.c_str());
  command = "mv '" + zipfile + "' '" + configuration->export_to_sword_module_path + "'";
  system (command.c_str());
  // Remove inputfile and directory created.
  unlink (inputfile.c_str());
  remove_directory (base_directory);
  // Notebook hides progress bar.
  progress->finish();
}
