/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "project.h"
#include "utilities.h"
#include "config.h"
#include <glib.h>
#include "directories.h"
#include "stylesheetutils.h"
#include "constants.h"
#include "gwrappers.h"


Project::Project (Configuration * configuration)
{
  projectname = configuration->project;
  myconfiguration = configuration;
  project_load ();
}


Project::Project (const ustring& project)
/*
Note: This second constructor is only for convenience. It is not as good as 
the first type, neither does it write changes to disc. But it is useful
when one only queries project settings, without editing them.
*/
{
  projectname = project;
  myconfiguration = NULL;
  project_load ();
}


Project::~Project ()
{
  if (myconfiguration) {
    // If the object goes out of scope, it saves the project settings.
    vector <ustring> output;
    output.push_back (VERSION);
    output.push_back ("There is no need for editing this file, because it is maintained by bibledit");
    output.push_back (editor_font);
    output.push_back (mystylesheet);
    output.push_back (versification);
    output.push_back (sword_description);
    output.push_back (sword_about);
    output.push_back (sword_lcsh);
    output.push_back (sword_license);
    output.push_back (sword_version);
    output.push_back (sword_language);
    try {
      write_lines (filename, output);
    }
    catch (exception & ex) { 
      cerr << ex.what () << endl;
    }
    // Save the printing fonts.
    try {
      write_lines (printing_fonts_filename, printing_fonts);
    }
    catch (exception & ex) { 
      cerr << ex.what () << endl;
    }
  }
}


void Project::project_load ()
{
  // Read the project file, if there is any. Add the bits that are missing.
  filename = gw_build_filename (directories_get_projects(), projectname, "projectsettings");
  ReadText rc (filename, true);
  for (unsigned int i = 0; i <= 1000; i++) {
    switch (i) {
      case 2:
      {
        if (i < rc.lines.size ()) {
          editor_font = rc.lines[i];
        }
        else {
          if (myconfiguration)
            editor_font = myconfiguration->editor_font_name;
        }
        break;
      }
      case 3:
      {
        if (i < rc.lines.size ()) {
          mystylesheet = rc.lines[i];
        }
        else
        {
          mystylesheet = "Standard";
        }
        break;
      }
      case 4:
      {
        if (i < rc.lines.size ()) {
          versification = rc.lines[i];
        }
        else
        {
          versification = "English";
        }
        break;
      }
      case 5:
      {
        if (i < rc.lines.size ())
          sword_description = rc.lines[i];
        else
          sword_description = "Bibledit project " + myconfiguration->project;
        break;
      }
      case 6:
      {
        if (i < rc.lines.size ())
          sword_about = rc.lines[i];
        else
          sword_about = sword_description;
        break;
      }
      case 7:
      {
        if (i < rc.lines.size ())
          sword_lcsh = rc.lines[i];
        else
          sword_lcsh = "Bible--Translation";
        break;
      }
      case 8:
      {
        if (i < rc.lines.size ())
          sword_license = rc.lines[i];
        else
          sword_license = "GNU General Public License";
        break;
      }
      case 9:
      {
        if (i < rc.lines.size ())
          sword_version = rc.lines[i];
        else
          sword_version = "1.0";
        break;
      }
      case 10:
      {
        if (i < rc.lines.size ())
          sword_language = rc.lines[i];
        else
          sword_language.clear();
        break;
      }

    }
  }
  // Read the printing fonts.
  printing_fonts_filename = gw_build_filename (directories_get_projects(), projectname, "printing-fonts");
  ReadText rc2 (printing_fonts_filename, true);
  printing_fonts.assign (rc2.lines.begin(), rc2.lines.end());
}


ustring Project::get_stylesheet ()
// Returns the stylesheet, checks it.
{
  ustring checked_sheet (mystylesheet);
  // See whether our stylesheet exists.
  vector<ustring> stylesheets;
  stylesheet_get_ones_available (stylesheets);
  set<ustring> sheets (stylesheets.begin(), stylesheets.end());
  // Sheet is there? Fine.
  if (sheets.find (checked_sheet) != sheets.end());
  // Sheets is not there - take Standard, if it's around.
  else if (sheets.find (STANDARDSHEET) != sheets.end())
    checked_sheet = STANDARDSHEET;
  // Else take first sheet in the list.
  else
    checked_sheet = stylesheets[0];
  // Update internal data.
  mystylesheet = checked_sheet;
  return checked_sheet;  
}


void Project::set_stylesheet (const ustring& stylesheet)
{
  mystylesheet = stylesheet;
}
