/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "stylesheetutils.h"
#include "directories.h"
#include <config.h>
#include "constants.h"
#include "xmlutils.h"
#include "gwrappers.h"
#include "gtkwrappers.h"


void stylesheet_get_ones_available (vector<ustring>& names)
// Gets the names of the stylesheets that are there.
{
  ReadDirectories rd (directories_get_stylesheets (), "", "");
  names.assign (rd.directories.begin(), rd.directories.end());
}


void stylesheet_create_new (const ustring& name)
// Create a new stylesheet from templates, named "name".
{
  ustring directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, name);
  create_directory (directory);
  ReadFiles rf (PACKAGE_DATA_DIR, "", ".style");
  for (unsigned int i = 0; i < rf.files.size(); i++) {
    ustring command;
    command = "cp ";
    ustring file = gw_build_filename (PACKAGE_DATA_DIR, rf.files[i]);
    command.append (file);
    command.append (" '");
    file = rf.files[i];
    file = file.substr (0, file.length() - 6);
    file.append (".xml");
    file = gw_build_filename (directory, file);
    command.append (file);
    command.append ("'");
    system(command.c_str());
  }
}


void stylesheet_delete (const ustring& name)
// Deletes a stylesheet.
{
  ustring directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, name);
  remove_directory (directory);
}


void stylesheet_copy (const ustring& from_name, const ustring& to_name)
// Copies one stylesheet to another.
{
  ustring from_dir;
  from_dir = directories_get_stylesheets ();
  from_dir = gw_build_filename (from_dir, from_name);
  ustring to_dir;
  to_dir = directories_get_stylesheets ();
  to_dir = gw_build_filename (to_dir, to_name);
  ustring command;
  command = "cp -r '";
  command.append (from_dir);
  command.append ("' '");
  command.append (to_dir);
  command.append ("'");
  system (command.c_str());
}


void stylesheet_export (const ustring& name, const ustring& filename)
// Exports stylesheet "name" to file "filename".
// The exported file is a zipped file that contains all the individual styles.
{
  ustring path = directories_get_stylesheets ();
  path = gw_build_filename (path, name);
  path = gw_shell_quote (path);
  ustring command;
  command = "cd ";
  command.append (path);
  command.append ("; zip '");
  command.append (filename + ".zip' ");
  command.append ("*.xml");
  system (command.c_str());  
}


ustring stylesheet_import (const ustring& filename)
// Imports a new stylesheet from "filename".
// It expects a file in the formst as it is given in the export function,
// that is a zipped file that contains all the individual styles.
{
  // Derive the name of the new stylesheet from the filename of the zipped archive.
  ustring name;
  name = gw_path_get_basename (filename);
  if (g_str_has_suffix (filename.c_str(), ".zip"))
    name.erase (name.length() - 4, 4);
  // Check whether it already exists.
  if (stylesheet_exists (name)) {
    gtkw_dialog_error (NULL, "A stylesheet with this name could not be imported, because it already exists");
    return "";
  }
  // Get the path of the new stylesheet.
  ustring path = directories_get_stylesheets ();
  path = gw_build_filename (path, name);
  // Create the directory.
  create_directory (path);
  path = gw_shell_quote (path);
  // Copy the zip archive there.
  ustring command;
  command = "cp '";
  command.append (filename);
  command.append ("' ");
  command.append (path);
  system (command.c_str());
  // Unpack the zip archive.
  command = "cd ";
  command.append (path);
  command.append ("; unzip *.zip; rm *.zip");
  system (command.c_str());
  // Return the name of the stylesheet we imported;
  return name;
}


bool stylesheet_exists (const ustring& name)
// Returns whether this stylesheet exists.
{
  vector<ustring> sheets;
  stylesheet_get_ones_available (sheets);
  set<ustring> existing_sheets (sheets.begin(), sheets.end());
  return (existing_sheets.find (name) != existing_sheets.end());
}


void stylesheet_get_styles (const ustring& stylesheet, vector<Style>& styles)
// Get all data of all stylesheets.
{
  ustring directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, stylesheet);
  ReadFiles rf (directory, "", ".xml");
  for (unsigned int i = 0; i < rf.files.size(); i++) {
    Style style (directory, rf.files[i].substr (0, rf.files[i].length() - 4), false);
    styles.push_back(style);
  }
}


void stylesheet_delete_style (const ustring& stylesheet, const ustring& marker)
// Deletes a style.
{
  ustring filename;
  filename = directories_get_stylesheets ();
  filename = gw_build_filename (filename, stylesheet, marker + ".xml");
  unlink (filename.c_str());  
}


void stylesheet_new_style (const ustring& stylesheet, const ustring& marker, const ustring& name, const ustring& info)
// Adds a new style.
{
  // Directory where to store the new marker.
  ustring directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, stylesheet);
  // See if the marker happens to be in the templates already.  
  ustring stylefile = gw_build_filename (PACKAGE_DATA_DIR, marker + ".style");
  if (g_file_test (stylefile.c_str(), G_FILE_TEST_IS_REGULAR)) {
    // In Templates: copy it.
    ustring new_stylefile = gw_build_filename (directory, marker);
    ustring command = "cp " + stylefile + " '" + new_stylefile + ".xml'";
    system (command.c_str());
  } else {
    // Unknown: save it with defaults.
    Style style (directory, marker, true);
    style.name = name;
    style.info = info;
  }
}


void stylesheet_save_style (const ustring& stylesheet, const ustring& marker,
                            const ustring& name, const ustring& info,
                            StyleType type, int subtype,
                            double fontsize, int fontpercentage,
                            const ustring& italic, const ustring& bold, 
                            const ustring& underline, const ustring& smallcaps,
                            bool superscript, const ustring& justification,
                            double spacebefore, double spaceafter,
                            double leftmargin, double rightmargin,
                            double firstlineindent, bool spancolumns,
                            bool userbool1, bool userbool2, bool userbool3,
                            int userint1, int userint2, int userint3,
                            ustring userstring1, ustring userstring2, ustring userstring3)
// Saves style information.
{
  ustring directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, stylesheet);
  Style style (directory, marker, true);
  style.name = name;
  style.info = info;
  style.type = type;
  style.subtype = subtype;
  style.fontsize = fontsize;
  style.fontpercentage = fontpercentage;
  style.italic = italic;
  style.bold = bold;
  style.underline = underline;
  style.smallcaps = smallcaps;
  style.superscript = superscript;
  style.justification = justification;
  style.spacebefore = spacebefore;
  style.spaceafter = spaceafter;
  style.leftmargin = leftmargin;
  style.rightmargin = rightmargin;
  style.firstlineindent = firstlineindent;
  style.spancolumns = spancolumns;
  style.userbool1 = userbool1;
  style.userbool2 = userbool2;
  style.userbool3 = userbool3;
  style.userint1 = userint1;
  style.userint2 = userint2;
  style.userint3 = userint3;
  style.userstring1 = userstring1;
  style.userstring2 = userstring2;
  style.userstring3 = userstring3;
}


int stylesheet_style_get_pointer (const vector<Style>& styles, const ustring& marker)
// Returns a pointer to "styles" which describes "marker".
// Or -1 if not found.
{
  // Search marker and return pointer.
  for (unsigned int i = 0; i < styles.size(); i++)
    if (marker == styles[i].marker)
      return i;
  // Ok, marker not found.
  return -1;
}


void stylesheet_initial_check ()
{
  // Get all available stylesheets.
  vector<ustring> stylesheets;
  stylesheet_get_ones_available (stylesheets);
  // No stylesheets available? Create standard one.
  if (stylesheets.size() == 0) {
    stylesheet_create_new (STANDARDSHEET);
    stylesheets.push_back (STANDARDSHEET);
  }
}


ustring stylesheet_internal_recently_used_filename (Configuration * configuration)
{
  return gw_build_filename (directories_get_stylesheets (), configuration->stylesheet, "recently-used");
}


vector<ustring> stylesheet_get_recently_used (Configuration * configuration)
{
  // Read all the styles (format: style <space> name).
  ustring filename = stylesheet_internal_recently_used_filename (configuration);
  ReadText rt (filename, true);
  // Take out the ones that are not in the stylesheet.
  vector<Style> styles;
  stylesheet_get_styles (configuration->stylesheet, styles);
  set<ustring> styles2;
  for (unsigned int i = 0; i < styles.size(); i++) {
    styles2.insert (styles[i].marker + " " + styles[i].name);
  }
  vector<ustring> existing_ones;
  for (unsigned int i = 0; i < rt.lines.size(); i++) {
    if (styles2.find (rt.lines[i]) != styles2.end())
      existing_ones.push_back (rt.lines[i]);
  }
  rt.lines.clear();
  rt.lines.assign (existing_ones.begin(), existing_ones.end());
  // Return result.
  return rt.lines;  
}


void stylesheet_set_recently_used (Configuration * configuration, vector<ustring> styles)
{
  try {
    ustring filename = stylesheet_internal_recently_used_filename (configuration);
    write_lines (filename, styles);
  }
  catch (exception & ex) { 
    cerr << ex.what () << endl;
  }
}


bool stylesheet_style_has_endmarker (const Style& style)
{
  bool endmarker = false;
  if (style.type == stInlineText)
    endmarker = true;
  if (style.type == stFootEndNote) {
    if (style.subtype == fentFootnote)
      endmarker = true;
    if (style.subtype == fentEndnote)
      endmarker = true;
    if (style.subtype == fentContentWithEndmarker)
      endmarker = true;
  }
  if (style.type == stCrossreference) {
    if (style.subtype == ctCrossreference)
      endmarker = true;
    if (style.subtype == ctContentWithEndmarker)
      endmarker = true;
  }
  return endmarker;
}
