/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "undo.h"


Undo::Undo (int dumy)
{
  clear ();
}


Undo::~Undo ()
{
}


void Undo::clear ()
{
  // Clear the whole undo stack.
  undo_position.clear ();
  undo_text.clear ();
  undo_insert.clear ();
  non_undoable_action = false;
  pointer = 0; // This is normally equal to the size of the stack.
}


void Undo::on_insert (GtkTextIter *pos, const ustring& text, int bytes)
{
/* It appears this handler is called _before_ the text has been inserted.
 * This appears from these two checks:
 * 1. The "modified" flag is still left as false when it was false.
 * 2. The text in the buffer is still left as it was.
 */
  stack_operation (gtk_text_iter_get_offset (pos), text, true);
}


void Undo::on_erase (GtkTextIter *range_begin, GtkTextIter *range_end)
{
/* It appears this handler is called _before_ the text has been deleted.
 * This appears from these two checks:
 * 1. The "modified" flag is still left as false when it was false.
 * 2. The text in the buffer is still left as it was.
 * In gtkmm this was different, and the handler was called after the text
 * was deleted. This made it necessary to use a shadow string that contained
 * the text of the buffer before the deletion. In Gtk+ this shadow string
 * could be removed. Another reason why the switch to Gtk+ was beneficial.
 */
  GtkTextBuffer *buffer = gtk_text_iter_get_buffer (range_begin);
  ustring text = gtk_text_buffer_get_text (buffer, range_begin, range_end, false);
  int offset = gtk_text_iter_get_offset (range_begin);
  stack_operation (offset, text, false);
}


void Undo::undo (GtkTextBuffer *buffer)
{
  non_undoable_action = true;
  if (can_undo ())
    {
      pointer--;
      if (undo_insert[pointer])
        {
          // Last action inserted, so we now erase it.
          /* Note that the position that has been saved, is the position in 
           * the textbuffer _before_ the actual insert was made.
           * So we should erase the text _starting_ from that position, 
           * and forward.
           */
          GtkTextIter range_begin;
          gtk_text_buffer_get_iter_at_offset (buffer, &range_begin, undo_position[pointer]);
          GtkTextIter range_end;
          gtk_text_buffer_get_iter_at_offset (buffer, &range_end, undo_position[pointer] + undo_text[pointer].length ());
          gtk_text_buffer_delete (buffer, &range_begin, &range_end);
          // Place cursor at the edit.
          gtk_text_buffer_place_cursor (buffer, &range_begin);
        }
      else
        {
          // Last action erased, so we now insert it.
          GtkTextIter position;
          gtk_text_buffer_get_iter_at_offset (buffer, &position, undo_position[pointer]);
          gtk_text_buffer_insert (buffer, &position, undo_text[pointer].c_str(), -1);
          // Place cursor at the edit.
          GtkTextIter cursor_position;
          gtk_text_buffer_get_iter_at_offset (buffer, &cursor_position, undo_position[pointer] + undo_text[pointer].length ());
          gtk_text_buffer_place_cursor (buffer, &cursor_position);
        }
    }
  non_undoable_action = false;
}


void Undo::redo (GtkTextBuffer *buffer)
{
  non_undoable_action = true;
  if (can_redo ())
    {
      if (undo_insert[pointer])
        {
          // Last action was an insert, so we insert the same data again.
          GtkTextIter position;
          gtk_text_buffer_get_iter_at_offset (buffer, &position, undo_position[pointer]);
          gtk_text_buffer_insert (buffer, &position, undo_text[pointer].c_str(), -1);
          // Place cursor at the edit.
          GtkTextIter cursor_position;
          gtk_text_buffer_get_iter_at_offset (buffer, &cursor_position, undo_position[pointer] + undo_text[pointer].length ());
          gtk_text_buffer_place_cursor (buffer, &cursor_position);
        }
      else
        {
          // Last action was an erase, so we erase the same data again.
          GtkTextIter range_begin;
          gtk_text_buffer_get_iter_at_offset (buffer, &range_begin, undo_position[pointer]);
          GtkTextIter range_end;
          gtk_text_buffer_get_iter_at_offset (buffer, &range_end, undo_position[pointer] + undo_text[pointer].length ());
          gtk_text_buffer_delete (buffer, &range_begin, &range_end);
          // Place cursor at the edit.
          GtkTextIter cursor_position;
          gtk_text_buffer_get_iter_at_offset (buffer, &cursor_position, undo_position[pointer]);
          gtk_text_buffer_place_cursor (buffer, &cursor_position);
        }
      pointer++;
    }
  non_undoable_action = false;
}


bool Undo::can_undo ()
{
  return (pointer > 0);
}


bool Undo::can_redo ()
{
  return (pointer < undo_insert.size ());
}


void Undo::stack_operation (int position, const ustring& text, bool insert)
{
  /*
     This stores the insert or erase operation in the buffer,
     and manages the buffer.
   */
  if (non_undoable_action)
    return;
  // If the pointer does not point to the end of the buffer, erase from there to the end.
  if (pointer != undo_insert.size ())
    {
      int n = undo_insert.size () - pointer;
      for (int i = 0; i < n; i++)
        {
          undo_position.pop_back ();
          undo_text.pop_back ();
          undo_insert.pop_back ();
        }
    }
  // Store the new data in the stack.
  undo_position.push_back (position);
  undo_text.push_back (text);
  undo_insert.push_back (insert);
  // We could be merging the current action with the most recent one,
  // if possible. This would be undoing / redoing whole words at once,
  // instead character by character. But this has not been implemented
  // at present, because I see little use for that. Perhaps later.

  // Trim buffer it if it grows too large.
  if (undo_insert.size () > 1000)
    {
      undo_position.erase (undo_position.begin ());
      undo_text.erase (undo_text.begin ());
      undo_insert.erase (undo_insert.begin ());
    }
  // Update the pointer.
  pointer = undo_insert.size ();
}

