/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include <libgen.h>
#include "windowscompanion.h"
#include <glib/gthread.h>
#include "bible.h"
#include "gwrappers.h"


#define STAGE_ZERO 0
#define STAGE_WAIT_CONNECT 1
#define STAGE_COMMUNICATE 2
#define STAGE_WAIT_RETRY 1000
#define STAGE_RETRY 1600


WindowsCompanion::WindowsCompanion (bool dummy)
{
  /*
     This controls a Bibledit Windows Companion program.
     This is done through connecting to a port on that program,
     and communicating with it.
   */
  sock = 0;
  Clear ();
}


WindowsCompanion::~WindowsCompanion ()
{
  Disconnect (false);
}


void WindowsCompanion::Clear ()
{
  run_stages = false;
  Disconnect ();
  stage = STAGE_ZERO;
  stages_running = false;
  connected = false;
}

void WindowsCompanion::Connect (const ustring & host, unsigned int portnumber)
{
  Clear ();
  hostname = host;
  port = portnumber;
  run_stages = true;
}


void WindowsCompanion::Tick ()
{
  if (run_stages)
    RunStages ();
}


void WindowsCompanion::Disconnect (bool message)
{
  if (sock) {
    // Send 'quit' command, close the socket.
    Sendln ("quit");
    close (sock);
    // Only write a message when the variabele is true.
    // This is done to prevent writing messages when 
    // Bibledit shuts down. This message would show up the next time
    // Bibledit starts. This would be confusing.
    if (message)
      log ("Disconnected");
  }
  sock = 0;
}


void WindowsCompanion::RunStages ()
{
  if (stages_running)
    return;
  stages_running = true;
  switch (stage)
  {
  case STAGE_ZERO:
    {
      /*
         Connecting might be slow. This will hang Bibledit.
         Therefore we use a thread to do this.
       */
      // Create a non-joinable thread -- it's deleted automatically on thread exit.
      // This sets up the connection.
      g_thread_create (GThreadFunc (local_thread_connect), gpointer (this), false, NULL);
      stage++;
      break;
    }
  case STAGE_WAIT_CONNECT:
    {
      // Wait for the thread to finish, then continue.
      if (thread_running)
        break;
      if (connected)
        {
          stage++;
        }
      else
        {
          stage = STAGE_WAIT_RETRY;
        }
      break;
    }
  case STAGE_COMMUNICATE:
    {
      break;
    }
  case STAGE_WAIT_RETRY:
    {
      stage++;
      break;
    }
  case STAGE_RETRY:
    {
      stage = STAGE_ZERO;
      break;
    }
  default:
    {
      stage++;
    }
  }
  stages_running = false;
}


void WindowsCompanion::thread_connect ()
{
  thread_running = true;
  // If the computer address can be converted to an IP, do so. 
  // If not, try to look it up in DNS.
  host = gethostbyname (hostname.c_str ());
  if (host)
  {
    // We got the host, now get a socket.
    if ((sock = socket (PF_INET, SOCK_STREAM, 0)) >= 0)
      {
        // We've got the socket, now set some variables.
        address.sin_family = AF_INET;
        address.sin_port = htons (port);
        // Take the first IP address associated with this hostname.
        memcpy (&address.sin_addr, host->h_addr_list[0],
                sizeof (address.sin_addr));
        // Now connect.
        if (connect (sock, (struct sockaddr *) &address, sizeof (address)))
          {
            log ("No connection possible");
            close (sock);
          }
        else
          {
            // We've a connection. Set the socket to non-blocking mode.
            fcntl (sock, F_SETFL, fcntl (sock, F_GETFL, 0) | O_NONBLOCK);
            log ("Connected");
            connected = true;
          }
      }
    else
      {
        log ("No socket available");
      }
  }
  else
  {
    /* We can't find an IP number */
    log ("Error looking up hostname");
  }
  thread_running = false;
}


void WindowsCompanion::log (const ustring & message)
{
  gw_message ("Windows Companion: " + message);
}


void WindowsCompanion::Sendln (const ustring & command)
{
  // Sends "command" to the windows companion.
  int result = 0;
  if (connected)
  {
    write (sock, command.c_str (), command.length ());
    result = write (sock, "\n", 1);
  }
  if (result < 0)
  {
    connected = false;
    log ("Error sending data");
    reconnect ();
  }
}


ustring WindowsCompanion::Readln ()
{
  /*
     Receives a line from the windows companion.
     If no data is available, or no complete line is available yet,
     it returns immediately with a null string.
   */

  if (sock)
  {
    char buf[1024];
    int len = 0;
    len = read (sock, buf, sizeof (buf));
    if (len > 0)
      {
        buf[len] = '\0';
        return trim (ustring (buf));
      }
    if (len < 0)
      {
        connected = false;
        log ("Error receiving data");
        close (sock);
        reconnect ();
      }
  }
  return "";
}


void WindowsCompanion::reconnect ()
{
  Disconnect (false);
  Connect (hostname, port);
}


void WindowsCompanion::local_thread_connect(gpointer data)
{
  ((WindowsCompanion *) data)->thread_connect();
};


void WindowsCompanion::sendreference (const ustring& book, const ustring& chapter, const ustring& verse)
{
  // As it does not respond properly to chapter/verses that are 0, care for that.
  unsigned int index = english_id_to_index (book);
  ustring goto_ch = chapter;
  if (goto_ch == "0") goto_ch = "1";
  ustring goto_vs = verse;
  if (goto_vs == "0") goto_vs = "1";
  ustring value = "gotoreference " + index_to_paratext_id (index) + " " + goto_ch + ":" + goto_vs;
  Sendln (value);
}
